/* $Id: e2_action.c,v 1.28 2004/07/06 00:08:43 cairo Exp $
 * 
 * Copyright (C) 2003-2004 tooar <tooar@gmx.net>
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file COPYING. If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/**
 * @file src/actions/e2_action.c 
 * @brief action system
 *
 * This file contains functions for the action system.
 **/

#include <stdio.h>
#include <stdlib.h>
#include "emelfm2.h"
#include "e2_action.h"
#include "e2_command_line.h"
#include "menu_actions_cb.h"
#include "debug.h"

//FIXME: temporary static array of actions, registration of actions will be
//moved to the modules they belong to
static E2_Action e2_actions[] =
{
	{ "<submenu>", E2_ACTION_TYPE_SUBMENU, NULL, FALSE, E2_ACTION_EXCLUDE_ACCEL  },
	{ "<separator>", E2_ACTION_TYPE_SEPARATOR, NULL, FALSE, E2_ACTION_EXCLUDE_ACCEL  },
	{ "<file actions>", E2_ACTION_TYPE_FILE_ACTIONS, NULL, FALSE, E2_ACTION_EXCLUDE_ACCEL | E2_ACTION_EXCLUDE_TOOLBAR },
	{ "<plugins>", E2_ACTION_TYPE_PLUGINS, NULL, FALSE, E2_ACTION_EXCLUDE_ACCEL },
	{ "<custom command>", E2_ACTION_TYPE_ITEM, e2_menu_action_custom_command_cb, TRUE, 0 },
	{ "file.open", E2_ACTION_TYPE_ITEM, open_cb, FALSE, 0 },
	{ "file.open in other", E2_ACTION_TYPE_ITEM, open_in_other_pane_cb, FALSE, 0 },
	{ "file.open with", E2_ACTION_TYPE_ITEM, open_with_cb, FALSE, 0 },
	{ "view", E2_ACTION_TYPE_ITEM, view_cb, FALSE, 0 },
	{ "quit", E2_ACTION_TYPE_ITEM, quit_cb, FALSE, 0 },
	{ "panes.select all", E2_ACTION_TYPE_ITEM, fileview_select_all, FALSE, 0 },
	{ "panes.toggle hidden", E2_ACTION_TYPE_ITEM, toggle_show_hidden, FALSE, 0 },
/*	{ "show context menu", E2_ACTION_TYPE_ITEM, show_menu_cb, FALSE, 0 },
	{ "show shift context menu", E2_ACTION_TYPE_ITEM, show_plugins_menu_cb, FALSE, 0 },
	{ "show ctrl context menu", E2_ACTION_TYPE_ITEM, show_user_menu_cb, FALSE, 0 }, */
	{ "find", E2_ACTION_TYPE_ITEM, find_cb, FALSE, 0 },
	{ "file.toggle tag", E2_ACTION_TYPE_ITEM, toggle_tag_cb, FALSE, 0 },
	{ "file.file info", E2_ACTION_TYPE_ITEM, file_info_cb, FALSE, 0 },
	{ "file.permission", E2_ACTION_TYPE_ITEM, permissions_cb, FALSE, 0 },
	{ "file.user/group", E2_ACTION_TYPE_ITEM, ownership_cb, FALSE, 0 },
	{ "file.rename", E2_ACTION_TYPE_ITEM, rename_cb, FALSE, 0 },
	{ "file.copy", E2_ACTION_TYPE_ITEM, copy_cb, FALSE, 0 },
	{ "file.copy as", E2_ACTION_TYPE_ITEM, copy_as_cb, FALSE, 0 },
	{ "file.move", E2_ACTION_TYPE_ITEM, move_cb, FALSE, 0 },
	{ "file.move as", E2_ACTION_TYPE_ITEM, move_as_cb, FALSE, 0 },
	{ "file.symlink", E2_ACTION_TYPE_ITEM, symlink_cb, FALSE, 0 },
	{ "file.symlink as", E2_ACTION_TYPE_ITEM, symlink_as_cb, FALSE, 0 },
	{ "file.delete", E2_ACTION_TYPE_ITEM, delete_cb, FALSE, 0 },
	{ "file.mkdir", E2_ACTION_TYPE_ITEM, e2_mkdir_dialog_show, FALSE, 0 },
	{ "config.configure", E2_ACTION_TYPE_ITEM, configure_cb, TRUE, 0 },
	{ "config.filetype", E2_ACTION_TYPE_ITEM, e2_menu_action_edit_filetype_cb, FALSE, 0 },
};
static gint e2_actions_size = sizeof (e2_actions) / sizeof (e2_actions[0]);

static GHashTable *actions_hash;
static GPtrArray *actions_array;
static GtkTreeStore *actions_store;

  /*********************/
 /***** callbacks *****/
/*********************/

static gint tree_compare_cb (GtkTreeModel *model, GtkTreeIter *a, GtkTreeIter *b,
	gpointer userdata)
{
	gint ret = 0;
	gchar *name1, *name2;
	gtk_tree_model_get(model, a, 1, &name1, -1);
	gtk_tree_model_get(model, b, 1, &name2, -1);
	if (name1 == NULL || name2 == NULL)
	{
		if (name1 == NULL && name2 == NULL)
		return ret;
		ret = (name1 == NULL) ? -1 : 1;
	} else
		ret = g_utf8_collate(name1,name2);
	g_free(name1);
	g_free(name2);
	return ret;
}

static gboolean visible_cb (GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	E2_ACTION_EXCLUDE ex1 = GPOINTER_TO_INT (data);
	E2_ACTION_EXCLUDE ex2;
	gtk_tree_model_get (model, iter, 2, &ex2, -1);
	return !(ex1 & ex2);
}

  /******************/
 /***** public *****/
/******************/

/**
 * @brief option helper for action columns
 *
 * This function can be used as source function for @a E2_OPTION_TREE_TYPE_SEL
 * columns of E2_OPTION_TYPE_TREE options.
 *
 * @param ex  exclude flags to select actions that should not be visible
 *
 * @return GtkTreeFilterModel of all actions
 **/
GtkTreeModel *e2_actions_fill_store (E2_ACTION_EXCLUDE ex)
{
	GtkTreeModel *filter = gtk_tree_model_filter_new (GTK_TREE_MODEL
		(actions_store), NULL);
	gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER (filter),
		(GtkTreeModelFilterVisibleFunc) visible_cb, GINT_TO_POINTER (ex),
		NULL);
	return filter;
}

E2_Action *e2_action_register (const gchar *name, E2_ACTION_TYPE type,
	gpointer func, gpointer data, gboolean has_arg)
{
	return _e2_action_register (name, type, func, data, has_arg, 0, NULL);
}

E2_Action *_e2_action_register (const gchar *name, E2_ACTION_TYPE type,
	gpointer func, gpointer data, gboolean has_arg, E2_ACTION_EXCLUDE exclude,
	gpointer data2)
{
	//unregister any previously registered actions with that name
	//(for convenience)
	e2_action_unregister (name);
	//now allocation the action structure and initialize it from the
	//function parameters
	E2_Action *action = g_malloc (sizeof (E2_Action));
	//freed when the action is removed from the actions_hash table
	//because the names are used as keys
	action->name = g_strdup (name);
	action->type = type;
	action->func = func;
	action->data = data;
	action->has_arg = has_arg;
	action->exclude = exclude;
	action->data2 = data2;

	//add the action to the static hash table and pointer array register
	g_hash_table_insert (actions_hash, action->name, action);
	g_ptr_array_add (actions_array, action);

	//to get structured lists, find the name of the parent action
	E2_Action *parent = NULL;
	gchar *search = g_strdup (name);
	gchar *child = g_utf8_strrchr (search, -1, '.');
	if (child == NULL)
		child = g_utf8_strrchr (search, -1, ':');
	if (child != NULL)
	{
		*child = '\0';
		parent = e2_action_get (search);
		if (parent == NULL)
			parent = e2_action_register (search, E2_ACTION_TYPE_DUMMY,
				NULL, NULL, FALSE);
	}

	//now add the action to the static tree store
	GtkTreeIter iter;
	if (parent == NULL)
		gtk_tree_store_append (actions_store, &iter, NULL);
	else
	{
		GtkTreeIter iter_parent;
		if (e2_tree_ref_to_iter (actions_store, parent->ref, &iter_parent))
			gtk_tree_store_append (actions_store, &iter, &iter_parent);
		else
			gtk_tree_store_append (actions_store, &iter, NULL);
	}
	//setup store; that's also convenience for the option system
	gtk_tree_store_set (actions_store, &iter, 0, action->name,
		1, child == NULL ? action->name : g_strstrip (child + 1),
		2, action->exclude, 3, action->type != E2_ACTION_TYPE_DUMMY, 4, TRUE, -1);

	//keep reference around to be able to to find out the parent iter
	//for possible children later on
	action->ref = e2_tree_iter_to_ref (actions_store, &iter);
	
	//cleanup
	if (child != NULL)
		*child = '.';
		g_free (search);
		
	return action;
}

gboolean e2_action_unregister (const gchar *name)
{
	E2_Action *action = g_hash_table_lookup (actions_hash, name);
	if (action == NULL)
		return FALSE;
	_e2_action_unregister (action);
	return TRUE;
}

void _e2_action_unregister (E2_Action *action)
{
	g_ptr_array_remove (actions_array, action);
	g_hash_table_remove (actions_hash, action->name);
	
	GtkTreeIter iter;
	if (e2_tree_ref_to_iter (actions_store, action->ref, &iter))
		gtk_tree_store_remove (actions_store, &iter);
	gtk_tree_row_reference_free (action->ref);
	g_free (action);
}

/**
 * @brief find an action by name
 *
 * The action with the name @a name is looked up in the internal hash table
 * and a pointer to the action object is returned. The name should be the
 * complete name of the action including any parent categories. If the action
 * cannot be found, NULL is returned.
 *
 * @param name  the name of the action
 *
 * @return  pointer to an action object or NULL
 **/
E2_Action *e2_action_get (gchar *name)
{
	E2_Action *action = g_hash_table_lookup (actions_hash, name);
	if (action == NULL)
	{
		gchar *name_try;
		if ((g_utf8_strchr (name, -1, '.')) == NULL)
			name_try = e2_utils_str_replace (name, ": ", ".");
		else
			name_try = e2_utils_str_replace (name, ".", ": ");
		action = g_hash_table_lookup (actions_hash, name_try);
		g_free (name_try);
	}
	return action;
}

/**
 * @brief run a certain action by name
 *
 * This function is to run a certain action and wraps around
 * e2_action_run_simple_from(). It sets the @a from widget
 * of e2_action_run_simple_from to NULL.
 *
 * @param name  the name of the action to run
 * @param arg  additional user data
 *
 * @return  TRUE on success, FALSE on failure
 **/
gboolean e2_action_run_simple (gchar *name, gpointer arg)
{
	return e2_action_run_simple_from (name, arg, NULL);
}

/**
 * @brief run a certain action by name
 *
 * This function is to run a certain action. First, the action specified by
 * @a name is looked up. Then a runtime object is created and the action is
 * run. Afterwards the runtime object is freed again. If the action could not
 * be found, an error message is printed to the output pane.
 * 
 * @param name  the name of the action to run
 * @param arg  additional user data
 * @param from  the widget the action is run from
 *
 * @return  TRUE on success, FALSE on failure
 **/
gboolean e2_action_run_simple_from (gchar *name, gpointer arg,
	GtkWidget *from)
{
	E2_Action *action = e2_action_get (name);
	if (action != NULL)
	{
		E2_ActionRuntime *actionrt = e2_action_pack_runtime (action,
			arg != NULL ? arg : NULL, NULL);
		e2_action_run_from (actionrt, from);
		e2_action_free_runtime (actionrt);
		return TRUE;
	} else
	{
		gchar *message = g_strconcat ("could not find action '", name, "'", NULL);
		e2_output_print (message, NULL, TRUE, FALSE, TRUE, NULL);
		g_free (message);
		return FALSE;
	}
}

/**
 * @brief  run a certain action; may be used as callback
 *
 * This function is to run a certain action and wraps e2_action_run_from().
 * It may be used as callback to Gtk+ signals. The first parameter @a widget
 * is ignored.
 * 
 * @param widget  this parameter is ignored
 * @param rt  the runtime object of the action
 *
 * @return  TRUE on success, FALSE on failure
 **/
gboolean e2_action_run (GtkWidget *widget, E2_ActionRuntime *rt)
{
	return e2_action_run_from (rt, widget);
}

/**
 * @brief  run a certain action
 *
 * This function is the real action run function. There are several wrappers
 * to simlify access. A runtime object has to be supplied that may contain
 * additional user data. The @a from widget is optional and may be used by
 * the action callback to determine on which widget to operate.
 *
 * @param rt  the runtime object from this action run
 * @param from  the widget the action is run from or NULL
 *
 * @return  TRUE on success, FALSE on failure
 **/
gboolean e2_action_run_from (E2_ActionRuntime *rt, GtkWidget *from)
{
	//some quick checks
	if (rt == NULL)
		return TRUE;
	if (rt->action->func == NULL)
		return TRUE;

	printd (DEBUG, "e2_action_run_from (rt:%s,from:)", rt->action->name);
	gboolean (*fun) (gpointer, gpointer, GtkWidget *) = rt->action->func;
	return fun (rt->action->data, rt->data, from);
}

/**
 * @brief create an action runtime object
 *
 * This function creates an action runtime object and initializes it with
 * supplied user data.
 *
 * @param action  the action to create the runtime object for
 * @param data  action user data
 * @param data_free  function to free data when the runtime object is freed
 *
 * @return newly allocated action runtime object that should be freed
 **/
E2_ActionRuntime *e2_action_pack_runtime (E2_Action *action, gpointer data,
	gpointer data_free)
{
	E2_ActionRuntime *rt = malloc (sizeof (E2_ActionRuntime));
	rt->action = action;
	rt->data = data;
	rt->data_free = data_free;
	return rt;
}

/**
 * @brief free action runtime object
 *
 * The runtime data is freed if a data free function has been supplied.
 * In addition the runtime object is freed and must not be used anymore after
 * calling this function.
 *
 * @param rt the action runtime object to free
 *
 * @return 
 **/
void e2_action_free_runtime (E2_ActionRuntime *rt)
{
	if (rt == NULL)
		return;

	if (rt->data_free != NULL)
	{
		void (*fun) (gpointer) = rt->data_free;
		fun (rt->data);
	}
	g_free (rt);
}

/**
 * @brief internal action system cleanup
 *
 * After calling the function, the action system may not be used anymore.
 * It is only called internally just before exit.
 *
 * @return 
 **/
void e2_actions_clean ()
{
	g_object_unref (actions_store);
	g_ptr_array_free (actions_array, TRUE);
	g_hash_table_destroy (actions_hash);
}

/**
 * @brief internal action system initialization
 *
 * This function is only called once at startup to initialize the action
 * system. Afterwars, it has no effect.
 * 
 * @return 
 **/
void e2_actions_init ()
{
	//assure that this function is only called once
	static gboolean init = FALSE;
	if (init)
		return;
	else
		init = TRUE;
	
	//initialize different local data structures that are used to keep track of
	//the action structures
	actions_hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
	actions_array = g_ptr_array_new ();
	actions_store = gtk_tree_store_new (5, G_TYPE_STRING, G_TYPE_STRING,
		G_TYPE_INT, G_TYPE_BOOLEAN, G_TYPE_BOOLEAN);
	//FIXME: is this still necessary?
	g_object_ref (actions_store);
	//store actions sorted in the tree store
	GtkTreeSortable *sortable = GTK_TREE_SORTABLE (actions_store);
    gtk_tree_sortable_set_sort_func (sortable, 0, tree_compare_cb, NULL,
		NULL);
    gtk_tree_sortable_set_sort_column_id (sortable, 0, GTK_SORT_ASCENDING);

	//FIXME: this should be removed at some point when all locally defined
	//actions have been moved to their modules
	gint i;
	for (i = 0; i < e2_actions_size; i++)
	{
		_e2_action_register (e2_actions[i].name, e2_actions[i].type,
			e2_actions[i].func, NULL, e2_actions[i].has_arg,
			e2_actions[i].exclude, NULL);
	}
}

