/* $Id: e2_complete.c,v 1.9 2004/08/14 12:01:08 cairo Exp $
 * 
 * Copyright (C) 2004 tooar <tooar@gmx.net>
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file COPYING. If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/**
 * @file src/command/complete/e2_complete.c 
 * @brief string auto completion
 *
 * This file contains functions to auto-complete strings, e.g. in a command
 * line or history. It has a registry which makes it possible to add new
 * completion checks. The type of the completion is then choosen from the
 * registry when completing strings.
 **/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "emelfm2.h"
#include "e2_complete.h"

static GList *checks = NULL;

  /*****************/
 /***** utils *****/
/*****************/

//return TRUE if the current E2_CompleteRuntime should be used
//to check the string
static gboolean should_check (GList *sel, E2_CompleteRuntime *rt)
{
	//FIXME: make this working
	return TRUE;
}

//find the word that should be completed on a line and return
//the string before the word and set up the word pointer correctly
//(pos is the current cursor position)
static gchar *find_word (const gchar *line, gint pos, gchar **word)
{
	//copy the string from the beginning of the line to the cursor position
	gchar *str = e2_utf8_ndup (line, pos);
	//quick exit in case the cursor is at the beginning of the line
	//word and return value are \0 strings
	if (pos == 0)
	{
		*word = g_strdup (str);
		return str;
	}
	//i is the character offset in str (line) where the word starts
	gint i;
	//search backwards from the currenct cusor position for the beginning
	//of the word
	for (i = pos; i > 0; i--)
	{
		gunichar cur = g_utf8_get_char (g_utf8_offset_to_pointer (str, i));
		gunichar before = g_utf8_get_char (g_utf8_offset_to_pointer (str, i - 1));
		//if the character before the space character is not backslash, we found
		//the beginning of the word and may break
		if ((cur == ' ') && (before != '\\'))
		{
			i++;
			break;
		}
	}
	//in case the word starts at the beginning of the line and the first character
	//is a space, increment i (this case is missed by the backward search above)
	if ((i == 0) && (g_utf8_get_char (str) == ' '))
		i++;
	//now
	gchar *tmp = g_strdup (g_utf8_offset_to_pointer (str, i));
	
	*word = e2_utf8_unescape (tmp, ' ');
	g_free (tmp);
	gchar *retval = e2_utf8_ndup (str, i);
	g_free (str);
	return retval;
}

  /******************/
 /***** public *****/
/******************/

gint e2_complete_str (gchar **complete, gint *pos, GList **found,
	E2_CompleteFlags flags, const gchar *name, ...)
{
	//run all checks
	gboolean all = FALSE;
	if ((name == NULL) || (g_str_equal (name, "all")))
		all = TRUE;
	GList *sel = NULL;
	if (!all)
	{
		va_list names;
		va_start (names, name);
		while (name)
		{
			sel = g_list_append (sel, (gpointer) name);
			name = va_arg (names, const gchar *);
		}
		va_end (names);
	}
	gchar *word;
	gchar *part1 = find_word (*complete, *pos, &word);

	gint count = 0;
	GList *tmp;
	for (tmp = checks; tmp != NULL; tmp = g_list_next (tmp))
	{
		if ((all) || (should_check (sel, tmp->data)))
		{
			E2_CompleteRuntime *rt = tmp->data;
			gint (* fun) (gchar *, gchar *, gint, GList **, E2_CompleteFlags *,
				gpointer) = rt->func;
			gint cur = fun (*complete, word, *pos, found, &flags, rt->data);
			count += cur;
			if (flags & E2_COMPLETE_FLAG_STOP)
				break;
		}
	}

	if (*found != NULL)
	{
		gchar *repl = NULL;
		gchar *part2 = g_utf8_offset_to_pointer (*complete, *pos);
		gint common = 0;
		if ((*found)->next != NULL)
		{
			repl = g_strdup ((*found)->data);
			gint len1 = g_utf8_strlen ((*found)->data, -1);
			gint common = len1;
			for (tmp = (*found)->next; tmp != NULL; tmp = g_list_next (tmp))
			{
				gint len2 = g_utf8_strlen (tmp->data, -1);
				gint min = MIN (len1, len2);
				if (common > min)
					common = min;
				gint i;
				for (i = 0; i < min; i++)
				{
					if ((((gchar *) (*found)->data)[i]) != (((gchar *) tmp->data)[i]))
						break;
				}
				if (common > i)
					common = i;
			}
			if (common == 0)
			{
				g_free (repl);
				repl = NULL;
			}
			else
				repl[common] = '\0';
		} else
		{
			repl = g_strdup ((*found)->data);
		} 
		if (repl != NULL)
		{
			gchar *tmp = e2_utf8_escape (repl, ' ');
			gchar *comp_new = g_strconcat (part1, tmp, part2, NULL);
			*pos += g_utf8_strlen (tmp, -1) - g_utf8_strlen (word, -1);
			g_free (tmp);
			g_free (*complete);
			*complete = comp_new;
			if (common != 0)
				repl[common] = ' ';
			g_free (repl);
		}
	}
	g_free (part1);
	g_free (word);
	return count;
}

void e2_complete_register_method (guint priority, const gchar *name,
	gchar *group, gchar *desc, gpointer func, gpointer data)
{
	E2_CompleteRuntime *rt = g_new0 (E2_CompleteRuntime, 1);
	rt->priority = priority;
	rt->name = g_strdup (name);
	rt->group = g_strdup (group);
	rt->desc = g_strdup (desc);
	rt->func = func;
	rt->data = data;
	checks = g_list_append (checks, rt);
}

/**
 * @brief initialize the auto completion system
 *
 * At startup, e2 calls this function internally to initialize the auto
 * completion system. It should not be called and does nothing after the
 * first run.
 *
 * @return 
 **/
void e2_complete_init ()
{
	RUN_ONCE_CHECK ();
	
	checks = NULL;

	//register default methods
	e2_complete_register_method (10, "executables from path", "default",
		_("search PATH for executable files"), e2_complete_path, NULL);
	e2_complete_register_method (100, "files & directories", "default",
		_("complete file and directory names"), e2_complete_files, NULL);
}
