/* $Id: e2_option__io.c,v 1.10 2004/05/23 02:10:18 cairo Exp $
 * 
 * Copyright (C) 2004 tooar <tooar@gmx.net>
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file COPYING. If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/**
 * @file e2_option__io.c
 * @brief config file functions
 *
 * This file contains functions to read and write config options from and to
 * file.
 **/

#include <stdlib.h>
#include <stdio.h>
#include <glib/gprintf.h>
#include "emelfm2.h"
#include "debug.h"
#include "build.h"
#include "e2_cl_option.h"

/**
 * @brief read config file @a fn
 *
 * First, the configuration is read from @a fn and parsed with
 * e2_option_read_from_string. If tree options have no configuration,
 * they are initialized with a default one.
 * (All other options have been intialized when they were registered.)
 *
 * @param fn filename inside the config directory
 */
void e2_option_read_from_file (gchar *fn)
{
	//find absolute path to config file
	gchar *filename = g_build_filename (e2_cl_options.config_dir, fn, NULL);
	printd (DEBUG, "option read from file '%s'", filename);
	gchar *cont;
	//get file contents
	if ((g_file_get_contents (filename, &cont, NULL, NULL)) == FALSE)
	{
		printd (WARN, "could not open config file '%s'", filename);
	} else
	{
		//everything goes into an array
		gchar **split = g_strsplit (cont, "\n", -1);
		//and now parse it
		e2_option_read_from_string (split);
		g_strfreev (split);
		g_free (cont);
	}
	g_free (filename);

	//check if any tree options need a default config
	gint i;
	for (i = 0; i < options_array->len; i++)
	{
		E2_OptionSet *set = g_ptr_array_index (options_array, i);
		if (set->type == E2_OPTION_TYPE_TREE)
			if ((!set->ex.tree.found_in_config_file) &&
				(set->ex.tree.def != NULL))
					//option needs default configuration which is
					//stored in ex.tree.def of the OptionSet
					e2_option_read_from_string (set->ex.tree.def);
	}
}

/**
 * @brief write current configuration to file
 *
 * The current configuration will be written to the default config file 'config'.
 */
void e2_option_write_to_file ()
{
	gchar *filename;
	filename = g_build_filename (e2_cl_options.config_dir, DEFAULT_CONFIG_FILE, NULL);
	printd (DEBUG, "write config file2: %s", filename);

	FILE *f;
	if ((f = fopen (filename, "w")) == NULL)
	{
		printd (WARN, "cannot write to config file: %s (%s)", filename, g_strerror (errno));
		return;
	}
	g_free (filename);

	g_fprintf 
	(
		f, "# " BINNAME " (v" VERSION ")\n\n"
		"# this is the " BINNAME " config file.\n"
		"# this is a generated file and will be overwritten automatically each run!\n\n"
		"# for tree options, you have to use \\| to escape |. besides you have to\n"
		"# escape < with \\< as well, if it is the first non-space character on the line.\n\n"
	);

	gint i;
	for (i = 0; i < options_array->len; i++)
	{
		E2_OptionSet *set = g_ptr_array_index (options_array, i);
		switch (set->type)
		{
			case E2_OPTION_TYPE_INT:
				g_fprintf (f, "# %s\n", set->desc);
				g_fprintf (f, "%s=%d\n", set->name, set->idef);
				break;
			case E2_OPTION_TYPE_BOOL:
				g_fprintf (f, "# %s\n", set->desc);
				g_fprintf (f, "%s=%s\n", set->name, set->sdef);
				break;
			case E2_OPTION_TYPE_STR:
			case E2_OPTION_TYPE_FONT:
			case E2_OPTION_TYPE_COLOR:
				g_fprintf (f, "# %s\n", set->desc);
				g_fprintf (f, "%s=%s\n", set->name, set->sdef);
				break;
			case E2_OPTION_TYPE_TREE:
				g_fprintf (f, "# %s\n", set->name);
				g_fprintf (f, "<%s\n", set->name);
				GtkTreeIter iter;
				if ((gtk_tree_model_get_iter_first (set->ex.tree.model, &iter)) == TRUE)
				{
					_e2_option_tree_write_to_file (f, set, &iter, 0);
				}
				g_fprintf (f, ">\n");
				break;
			case E2_OPTION_TYPE_SEL:
				g_fprintf (f, "# %s\n", set->desc);
				g_fprintf (f, "%s=%s\n", set->name, set->sdef);
				break;
			default:
				printd (WARN, "don't know how to write option '%s' to config file",
					set->name);
				break;
		}
	}

	fclose (f);
}
