/* $Id: e2_option_tree.c,v 1.27 2004/08/14 12:01:08 cairo Exp $
 * 
 * Copyright (C) 2003-2004 tooar <tooar@gmx.net>
 * 
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file COPYING. If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/**
 * @file src/config/e2_option_tree.c 
 * @brief functions to handle tree options
 *
 * functions to handle tree options
 *
 **/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <glib/gprintf.h>
#include "emelfm2.h"
#include "e2_option.h"
#include "e2_option_tree.h"
#include "debug.h"
#include "mg-cell-renderer-list.h"
#include "eggcellrendererkeys.h"
#include "e2_dialog.h"
#include "e2_button.h"

static GtkWidget *config_dialog = NULL;

  /**************************/
 /***** data functions *****/
/**************************/

static void bool_data_func (GtkTreeViewColumn *tree_column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	E2_OptionTreeSet *set)
{
	gint *column = g_object_get_data (G_OBJECT (cell), "column");
	
	gboolean b;
	gtk_tree_model_get (model, iter, (int) column, &b, -1);
	
	gboolean visible = TRUE;
	if (set->visible_check_func != NULL)
	{
		gboolean (*fun) (GtkTreeModel *, GtkTreeIter *, GtkCellRenderer *,
			gpointer) = set->visible_check_func;
		visible = fun (model, iter, cell,  set->visible_check_func_data);
	}
	g_object_set (G_OBJECT (cell), "visible", visible, "active", b, NULL);
}

static void str_data_func (GtkTreeViewColumn *tree_column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	E2_OptionTreeSet *set)
{
	gint *column = g_object_get_data (G_OBJECT (cell), "column");
	
	gchar *str;
	gtk_tree_model_get (model, iter, (int) column, &str, -1);
	
	gboolean visible = TRUE;
	if (set->visible_check_func != NULL)
	{
		gboolean (*fun) (GtkTreeModel *, GtkTreeIter *, GtkCellRenderer *,
			gpointer) = set->visible_check_func;
		visible = fun (model, iter, cell,  set->visible_check_func_data);
	}
	g_object_set (G_OBJECT (cell), "visible", visible, "text", str, NULL);
	g_free (str);
}

static void tree_view_sel_data_func (GtkTreeViewColumn *tree_column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	E2_OptionTreeSet *set)
{
	GtkTreePath *path = gtk_tree_model_get_path (model, iter);
	gchar *name;
	gint *column = g_object_get_data (G_OBJECT (cell), "column");
	gtk_tree_model_get (model, iter, column, &name, -1);
	gboolean visible = TRUE;
	if (set->visible_check_func != NULL)
	{
		gboolean (*fun) (GtkTreeModel *, GtkTreeIter *, GtkCellRenderer *,
			gpointer) = set->visible_check_func;
		visible = fun (model, iter, cell,  set->visible_check_func_data);
	}
	g_object_set (cell, "text", name, "visible", visible, NULL);
	g_free (name);
	gtk_tree_path_free (path);
}

static void key_data_func (GtkTreeViewColumn *tree_column,
	GtkCellRenderer *cell, GtkTreeModel *model, GtkTreeIter *iter,
	E2_OptionTreeSet *set)
{
	gint *column = g_object_get_data (G_OBJECT (cell), "column");
	
	gchar *key;
	gtk_tree_model_get (model, iter, (int) column, &key, -1);
	
	GdkModifierType mask = 0;
	guint keyval = 0;
	if (key != NULL)
	{
		gtk_accelerator_parse (key, &keyval, &mask);
	}
	gboolean visible = TRUE;
	if (set->visible_check_func != NULL)
	{
		gboolean (*fun) (GtkTreeModel *, GtkTreeIter *, GtkCellRenderer *,
			gpointer) = set->visible_check_func;
		visible = fun (model, iter, cell,  set->visible_check_func_data);
	}
	g_object_set (G_OBJECT (cell), "visible", visible, "accel_key", keyval,
		"accel_mask", mask, NULL);
}

  /*********************/
 /***** callbacks *****/
/*********************/

static gboolean draggable_check_cb (GtkTreeDragSource *drag_source,
	GtkTreePath *path)
{
	if (!GTK_IS_TREE_MODEL (drag_source))
		return TRUE;
	GtkTreeModel *model = GTK_TREE_MODEL (drag_source);
	E2_OptionSet *set = g_object_get_data (G_OBJECT (model), "e2-option-set");
	if (set == NULL)
		return TRUE;
	if (set->ex.tree.draggable_check_func != NULL)
	{
		gboolean (*fun) (GtkTreeDragSource *, GtkTreePath *) =
			set->ex.tree.draggable_check_func;
		return fun (drag_source, path);
	}
	else
		return TRUE;
}

static void key_edited_cb (GtkCellRendererText *cell, const char *path_string,
	guint keyval, GdkModifierType mask, guint hardware_keycode,
	E2_OptionSet *set)
{
	GtkTreePath *path = gtk_tree_path_new_from_string (path_string);
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter (set->ex.tree.model, &iter, path))
	{
		gint *column = g_object_get_data (G_OBJECT (cell), "column");
		gchar *key = gtk_accelerator_name (keyval, mask);
		_e2_option_tree_set (set, &iter, (gint) column, key);
		g_free (key);
	}
	gtk_tree_path_free (path);
}

static gboolean key_press_cb (GtkWidget *widget, GdkEventKey *event, gpointer data)
{
	if (event->keyval == GDK_Return)
		return e2_cd_ok_cb (NULL, FALSE);
	if (event->keyval == GDK_Escape)
		return e2_cd_cancel_cb (NULL);
	return FALSE;
}

static gboolean tree_view_up_cb (GtkWidget *widget, E2_OptionSet *set)
{
	GtkTreeModel *model;
	GtkTreeIter iter;

	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (set->widget)), &model, &iter))
	{
		GtkTreePath *path;
		path = gtk_tree_model_get_path (model, &iter);
		if ((gtk_tree_path_prev (path)) == TRUE)
		{
			_e2_option_tree_move (set, &iter, path, TRUE, TRUE);
		}
		if (path)
	    	gtk_tree_path_free (path);
	}
	return TRUE;
}

static gboolean tree_view_down_cb (GtkWidget *widgget, E2_OptionSet *set)
{
	GtkTreeModel *model;
	GtkTreeIter iter;

	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (set->widget)), &model, &iter))
	{
		GtkTreePath *path;
		path = gtk_tree_model_get_path (model, &iter);
		gtk_tree_path_next (path);
		_e2_option_tree_move (set, &iter, path, TRUE, FALSE);
		if (path)
	    	gtk_tree_path_free (path);
	}
	return TRUE;
}

static void treeview_cell_pixbuf_data_func (GtkTreeViewColumn *tree_column, GtkCellRenderer *cell,
	GtkTreeModel *model, GtkTreeIter *iter, gpointer data)
{
	gint *column = g_object_get_data (G_OBJECT (tree_column), "column");
	gchar *icon;
	gtk_tree_model_get (model, iter, (int) column, &icon, -1);
	GtkStockItem item;
	if ((gtk_stock_lookup (icon, &item)) == TRUE)
	{
		g_object_set (GTK_CELL_RENDERER (cell), "pixbuf", NULL, NULL);
		g_object_set (GTK_CELL_RENDERER (cell), "stock-id", icon, NULL);
	} else
	{
		GdkPixbuf *pixbuf;
		GdkPixbuf *pixbuf2;
		if ((pixbuf = gdk_pixbuf_new_from_file (icon, NULL)) != NULL)
		{
			pixbuf2 = pixbuf;
			pixbuf = gdk_pixbuf_scale_simple (pixbuf2, 16, 16, GDK_INTERP_BILINEAR);
			g_object_unref (pixbuf2);
		}
		g_object_set (GTK_CELL_RENDERER (cell), "stock-id", NULL, NULL);
  		g_object_set (GTK_CELL_RENDERER (cell), "pixbuf", pixbuf, NULL);
		if (pixbuf) g_object_unref (pixbuf);
	} 
	g_free (icon);
}

static void treeview_icon_column_set_cb (GtkWidget *dialog, gint response, E2_OptionSet *set) 
{
	if ((response == GTK_RESPONSE_APPLY) || (response == GTK_RESPONSE_OK) ||
		(response == 110))
	{
		GtkTreeIter iter;
		if ((gtk_tree_selection_get_selected (gtk_tree_view_get_selection
			(GTK_TREE_VIEW (set->widget)), NULL, &iter)) == TRUE)
		{
			gint *column = g_object_get_data (G_OBJECT (dialog), "column");
			gchar *icon = g_object_get_data (G_OBJECT (dialog), "image");
			_e2_option_tree_set (set, &iter, (gint) column, icon);
		}
	}
	if ((response != GTK_RESPONSE_APPLY) && (response != 110))
		gtk_widget_destroy (dialog);
}

static void activated_cb (GtkTreeView *treeview, GtkTreePath *path,
	GtkTreeViewColumn *tree_column, E2_OptionSet *set)
{
	GtkWidget *dialog = NULL;
	void *column = g_object_get_data (G_OBJECT (tree_column), "column");
	GList *list = g_list_nth (set->ex.tree.columns, (gint) column);
	E2_OptionTreeSet *opt = list->data;
	if (opt->type == E2_OPTION_TREE_TYPE_ICON)
	{
		//get the current value
		gchar *icon;
		GtkTreeIter iter;
		gtk_tree_model_get_iter (GTK_TREE_MODEL (set->ex.tree.model), &iter, path);
		gtk_tree_model_get (set->ex.tree.model, &iter, (gint) column, &icon, -1);
		gchar *dialog_name = g_strconcat ("select icon for ", set->name, NULL);
		dialog = e2_sid_create (config_dialog, dialog_name, icon,
			treeview_icon_column_set_cb, set);
		g_free (dialog_name);
		g_object_set_data (G_OBJECT (dialog), "column", column);
	}
}
static gboolean edited_cb (GtkCellRendererText *cell, 
	const gchar *path_string, const gchar *new_text, E2_OptionSet *set)
{
	//find out where
	GtkTreePath *path = gtk_tree_path_new_from_string (path_string);
	gint *column;
	column = g_object_get_data (G_OBJECT (cell), "column");
	GtkTreeIter iter;
	gtk_tree_model_get_iter (GTK_TREE_MODEL (set->ex.tree.model), &iter, path);

	//change model
	gtk_tree_store_set (GTK_TREE_STORE (set->ex.tree.model), &iter, column, new_text, -1);

	//clean up
	gtk_tree_path_free (path);
	return TRUE;
}

static gboolean list_edited_cb (MgCellRendererList *cell, 
	gchar *path_string, gchar *new_text, E2_OptionSet *set)
{
	//find out where
	GtkTreePath *path = gtk_tree_path_new_from_string (path_string);
	gint *column;
	column = g_object_get_data (G_OBJECT (cell), "column");
	GtkTreeIter iter;
	gtk_tree_model_get_iter (GTK_TREE_MODEL (set->ex.tree.model), &iter, path);

	GtkTreeIter iter_sel;
	GtkTreeModel *model_sel;
	if (gtk_tree_selection_get_selected (cell->selection,
		&model_sel, &iter_sel))
	{
		gchar *cur;
		gtk_tree_model_get (model_sel, &iter_sel, 0, &cur, -1);
	
		//change model
		gtk_tree_store_set (GTK_TREE_STORE (set->ex.tree.model), &iter, column, cur, -1);
	}

	gtk_widget_grab_focus (set->widget);

	//clean up
	gtk_tree_path_free (path);
	return TRUE;
}

//callback after tree view cell was toggled
static gboolean toggle_cb (GtkCellRendererToggle *cell,
	const gchar *path_str, E2_OptionSet *set)
{
	//find out where
	GtkTreePath *path = gtk_tree_path_new_from_string (path_str);
	GtkTreeIter iter;
	gint *column;
	column = g_object_get_data (G_OBJECT (cell), "column");
	gtk_tree_model_get_iter (set->ex.tree.model, &iter, path);
	
	//reverse current value
	gboolean toggle_item;
	gtk_tree_model_get (set->ex.tree.model, &iter, column, &toggle_item, -1);
	toggle_item ^= 1;

	//change model
	gtk_tree_store_set (GTK_TREE_STORE (set->ex.tree.model), &iter, column, toggle_item, -1);

	//clean up
	gtk_tree_path_free (path);
	return TRUE;
}

static void e2_cd_option_show_popup_cb (MgCellRendererList *cell,
	const gchar *path_string, gint x1, gint y1, gint x2, gint y2, E2_OptionSet *set)
{
	if (cell->model == NULL)
	{
		void *column = g_object_get_data (G_OBJECT (cell), "column");
		GList *list = g_list_nth (set->ex.tree.columns, (gint) column);
		E2_OptionTreeSet *opt = list->data;
		GtkTreeModel *(*fun) (gpointer) = opt->extra;
		cell->model = fun (opt->extra_data);
	}
}

static void e2_cd_option_hide_popup_cb (MgCellRendererList *cell,
	GtkWidget *widget)
{
	gtk_widget_grab_focus (widget);
}

static gboolean tree_view_del_cb (GtkWidget *button, E2_OptionSet *set)
{
	GtkTreeIter iter;
	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection
			(GTK_TREE_VIEW (set->widget)), NULL, &iter))
	{
		GtkTreePath *path = gtk_tree_model_get_path (set->ex.tree.model, &iter);

		gint response = GTK_RESPONSE_YES;
		gint children = gtk_tree_model_iter_n_children (set->ex.tree.model, &iter);
		if (children > 0)
		{
			gchar *question = g_strdup_printf ("are you sure that you want to "
				"delete this entry and <b>%d %s</b>?", children,
				children == 1 ? "child" : "children");
			
			GtkWidget *dialog = e2_dialog_create (GTK_STOCK_DIALOG_QUESTION, question, "confirm row delete", NULL, NULL);
			response = e2_dialog_show (dialog, config_dialog, E2_BUTTON_NO | E2_BUTTON_YES,
				E2_DIALOG_RUN | E2_DIALOG_FREE);
			g_free (question);
		}

		if (response == GTK_RESPONSE_YES)
			_e2_option_tree_del (set, &iter);

		if (iter.user_data == NULL)
		{
			if (!gtk_tree_path_prev (path))
				if (!gtk_tree_path_up (path))
					path = NULL;
		} else
			path = gtk_tree_model_get_path
				(GTK_TREE_MODEL (set->ex.tree.model), &iter);
		if ((path != NULL) &&
			(gtk_tree_model_get_iter_first (set->ex.tree.model, &iter)))
		{
			gtk_tree_view_set_cursor (GTK_TREE_VIEW (set->widget), path,
				gtk_tree_view_get_column (GTK_TREE_VIEW (set->widget), 0),
				FALSE);
		}
	}
	return TRUE;
}

static gboolean tree_view_add_child_cb (GtkWidget *wid, E2_OptionSet *set)
{
	GtkTreeIter iter;
	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (set->widget)), NULL, &iter))
	{
		if (iter.user_data != NULL)
		{
			GtkTreeIter iter2;
			_e2_option_tree_add_default (set, &iter2, &iter, FALSE);
		}
	} else
	{
		_e2_option_tree_add_default (set, &iter, NULL, FALSE);
	}

	return TRUE;
}

static gboolean tree_view_add_below_cb (GtkWidget *wid, E2_OptionSet *set)
{
	GtkTreeIter iter;
	if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection (GTK_TREE_VIEW (set->widget)), NULL, &iter))
	{
		if (iter.user_data != NULL)
		{
			GtkTreeIter iter2;
			_e2_option_tree_add_default (set, &iter2, &iter, TRUE);
		}
	} else
	{
		_e2_option_tree_add_default (set, &iter, NULL, TRUE);
	}

	return TRUE;
}

  /******************/
 /***** public *****/
/******************/

void e2_option_tree_add_widget (GtkWidget *parent, GtkWidget *box,
	E2_OptionSet *set)
{
	config_dialog = parent;
	GtkWidget *vbox = gtk_vbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (box), vbox, TRUE, TRUE, 0);
	GtkWidget *sw = e2_widget_add_sw (vbox, GTK_POLICY_AUTOMATIC,
		GTK_POLICY_AUTOMATIC, TRUE, 5);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (sw), GTK_SHADOW_IN);
	
	//setup treeview
	GtkTreeView *treeview = GTK_TREE_VIEW (gtk_tree_view_new_with_model
		((GtkTreeModel *) set->ex.tree.model));
	gtk_container_add (GTK_CONTAINER (sw), GTK_WIDGET (treeview));
	gtk_tree_view_set_reorderable (treeview, TRUE);
	gtk_tree_view_set_headers_clickable (treeview, FALSE);
	gtk_tree_view_set_rules_hint (treeview, FALSE);

	set->widget = GTK_WIDGET (treeview);

	GtkTreeSelection *selection = gtk_tree_view_get_selection (treeview);
	gtk_tree_selection_set_mode (selection , GTK_SELECTION_BROWSE);
	g_object_set_data (G_OBJECT (set->ex.tree.model), "e2-option-set", set);
//	if (set->ex.tree.draggable_check_func != NULL)
//	{
//		printd (INFO, "drag func: %d", set->ex.tree.draggable_check_func);
		GtkTreeDragSourceIface *iface = GTK_TREE_DRAG_SOURCE_GET_IFACE (set->ex.tree.model);
//		iface->row_draggable = set->ex.tree.draggable_check_func;
		iface->row_draggable = draggable_check_cb;
//	}
	if (set->ex.tree.selection_check_func != NULL)
		gtk_tree_selection_set_select_function (selection, (GtkTreeSelectionFunc)
			set->ex.tree.selection_check_func, set, NULL);

	//connect callbacks
	//no longer needed as it seems
//	g_signal_connect (treeview, "drag-drop", G_CALLBACK (drop_cb), set);
//	g_signal_connect (treeview, "drag-end", G_CALLBACK (drag_end_cb), set);
	g_signal_connect_after (treeview, "key_press_event", G_CALLBACK (key_press_cb),
		NULL);
	g_signal_connect (treeview, "row-activated", G_CALLBACK (activated_cb),
		set);

	//add columns
	int j = 0;
	GList *collist;
	for (collist = set->ex.tree.columns; collist != NULL; collist = g_list_next (collist))
	{
		GtkCellRenderer *renderer;
		GtkTreeViewColumn *column;
		E2_OptionTreeSet *opt = collist->data;
		gboolean editable = !(opt->flags & E2_OPTION_TREE_SET_NOT_EDITABLE);
		switch (opt->type)
		{
			case E2_OPTION_TREE_TYPE_BOOL:
				renderer = gtk_cell_renderer_toggle_new ();
				g_object_set (renderer, "activatable", TRUE, NULL);
						g_object_set_data (G_OBJECT (renderer), "column", (gint *) j);
				g_signal_connect (renderer, "toggled",
					G_CALLBACK (toggle_cb), set);
				column = gtk_tree_view_column_new_with_attributes
					(opt->name, renderer, "active", j, NULL);
				g_object_set (column, "alignment", 0.5, NULL);
				gtk_tree_view_column_set_resizable (GTK_TREE_VIEW_COLUMN (column), TRUE);
				gtk_tree_view_column_set_cell_data_func (column, renderer,
					(GtkTreeCellDataFunc) bool_data_func, opt, NULL);
				gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
				break;
			case E2_OPTION_TREE_TYPE_STR:
				renderer = gtk_cell_renderer_text_new ();
				g_object_set (renderer, "editable", editable, NULL);
						g_object_set_data (G_OBJECT (renderer), "column", (gint *) j);
				g_signal_connect (renderer, "edited",
					G_CALLBACK (edited_cb), set);
				column = gtk_tree_view_column_new_with_attributes
					(opt->name, renderer, "text", j, NULL);
				gtk_tree_view_column_set_sizing (GTK_TREE_VIEW_COLUMN (column),
					GTK_TREE_VIEW_COLUMN_GROW_ONLY);
				gtk_tree_view_column_set_resizable (GTK_TREE_VIEW_COLUMN (column), TRUE);
				gtk_tree_view_column_set_cell_data_func (column, renderer,
					(GtkTreeCellDataFunc) str_data_func, opt, NULL);
				gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
				break;
			case E2_OPTION_TREE_TYPE_ICON:
				renderer = gtk_cell_renderer_pixbuf_new ();
				g_object_set (renderer, "stock-size", GTK_ICON_SIZE_MENU, NULL);
				column = gtk_tree_view_column_new_with_attributes
					(opt->name, renderer, NULL);
				gtk_tree_view_column_set_cell_data_func (column, renderer,
					treeview_cell_pixbuf_data_func, NULL, NULL);
				g_object_set_data (G_OBJECT (renderer), "column", (gint *) j);
				g_object_set_data (G_OBJECT (column), "column", (gint *) j);
				gtk_tree_view_column_set_sizing (GTK_TREE_VIEW_COLUMN (column),
					GTK_TREE_VIEW_COLUMN_GROW_ONLY);
				gtk_tree_view_column_set_resizable (GTK_TREE_VIEW_COLUMN (column), TRUE);
				gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
				break;
			case E2_OPTION_TREE_TYPE_SEL:
				renderer = mg_cell_renderer_list_new ();
				g_object_set (renderer, "editable", editable, NULL);
				g_object_set_data (G_OBJECT (renderer), "column", (gint *) j);
				//FIXME: bad hack for the cell renderer to be able to calculate
				//correct values if activated through a key-press
				g_object_set_data (G_OBJECT (renderer), "scrolled-window", sw);
				g_signal_connect (renderer, "edited",
					G_CALLBACK (list_edited_cb), set);
				g_signal_connect (renderer, "show-popup",
					G_CALLBACK (e2_cd_option_show_popup_cb), set);
				g_signal_connect (renderer, "hide-popup",
					G_CALLBACK (e2_cd_option_hide_popup_cb), treeview);
				column = gtk_tree_view_column_new_with_attributes
					(opt->name, renderer, NULL);
				gtk_tree_view_column_set_sizing (GTK_TREE_VIEW_COLUMN (column),
					GTK_TREE_VIEW_COLUMN_GROW_ONLY);
				gtk_tree_view_column_set_resizable (GTK_TREE_VIEW_COLUMN (column), TRUE);
				gtk_tree_view_column_set_cell_data_func (column, renderer, 
					(GtkTreeCellDataFunc) tree_view_sel_data_func, opt, NULL);
///				g_object_set_data (G_OBJECT (column),
//					"filter-func", tree_view_sel_data_func);
				gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
				break;
			case E2_OPTION_TREE_TYPE_KEY:
				renderer = egg_cell_renderer_keys_new ();
				g_object_set (renderer, "editable", TRUE, NULL);
				if (!editable)
					g_object_set (renderer, "visible", FALSE, NULL);
				g_object_set_data (G_OBJECT (renderer), "column", (gint *) j);
				g_signal_connect (G_OBJECT (renderer), "keys_edited",
					G_CALLBACK (key_edited_cb), set);

				column = gtk_tree_view_column_new_with_attributes
					(opt->name, renderer, NULL);
				gtk_tree_view_column_set_resizable (GTK_TREE_VIEW_COLUMN (column), TRUE);
				gtk_tree_view_column_set_cell_data_func (column, renderer,
					(GtkTreeCellDataFunc) key_data_func, opt, NULL);
				gtk_tree_view_append_column (GTK_TREE_VIEW (treeview), column);
				break;
			default:
				printd (WARN, "don't know how to render column %d of %s", j, set->name);
				break;
		}
		j++;
	}
	
	//spacer column
	if ((int) set->ex.tree.flags & E2_OPTION_TREE_LAST_COL_EMPTY)
	{
		GtkCellRenderer *renderer = gtk_cell_renderer_text_new ();
		GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes
			("", renderer, NULL);
		gtk_tree_view_append_column (treeview, column);
	}
	
	//now expand everything
	gtk_tree_view_expand_all (treeview);
	
	//add control buttons
	GtkWidget *bbox = gtk_hbox_new (FALSE, 5);
	gtk_box_pack_start (GTK_BOX (vbox), bbox, FALSE, FALSE, 0);
	E2_Quintet *q = e2_utils_quintet_new ();
	g_object_set_data_full (G_OBJECT (treeview), "e2-config-buttons", q, g_free);
	if (set->ex.tree.flags & E2_OPTION_TREE_UP_DOWN)
	{
		q->a = e2_button_add (bbox, FALSE, 0, "_up", GTK_STOCK_GO_UP,
			"move the selected row one place up",
			tree_view_up_cb, set);
		q->b = e2_button_add (bbox, FALSE, 0, "_down", GTK_STOCK_GO_DOWN,
			"move the selected row one place down",
			tree_view_down_cb, set);
	}
	if (set->ex.tree.flags & E2_OPTION_TREE_ADD_DEL)
	{
		q->c = e2_button_add (bbox, FALSE, 0, "_remove", GTK_STOCK_DELETE,
			"remove the selected row",  tree_view_del_cb, set);
		q->d = e2_button_add (bbox, FALSE, 0, "_add", GTK_STOCK_ADD,
			"add a row after the currently selected row in the tree",
			tree_view_add_below_cb, set);	
		if (!((gint) set->ex.tree.flags & E2_OPTION_TREE_LIST))
			q->e = e2_button_add (bbox, FALSE, 0, "add c_hild", GTK_STOCK_ADD,
				"add a child row to the currently selected one",
				tree_view_add_child_cb, set);	
	}

	//select first row
	GtkTreeIter iter;
	if (gtk_tree_model_get_iter_first (GTK_TREE_MODEL (set->ex.tree.model), &iter))
	{
		GtkTreePath *path = gtk_tree_model_get_path (GTK_TREE_MODEL
			(set->ex.tree.model), &iter);
		gtk_tree_view_set_cursor (treeview, path,
			gtk_tree_view_get_column (treeview, 0), FALSE);
		gtk_tree_path_free (path);
	}
}

E2_OptionSet *e2_option_tree_register (gchar *name, gchar *group, gchar *desc,
	gchar *depends, gpointer selection_check_func, gpointer draggable_check_func,
	E2_OptionTypeTreeFlags flags)
{
	E2_OptionSet *set;
	if ((set = _e2_option_get (name)) != NULL)
		return set;
	set = malloc (sizeof (E2_OptionSet));
	//set up E2_OptionSet
	set->name = name;
	set->group = group;
	set->type = E2_OPTION_TYPE_TREE;
	set->desc = desc;
	set->tip = desc;
	set->depends = depends;
	set->ex.tree.def = NULL;
	set->ex.tree.columns_num = 0;
	set->ex.tree.columns = NULL;
	set->ex.tree.flags = flags;
	set->ex.tree.model = gtk_tree_store_new (1, G_TYPE_INT);
	set->ex.tree.found_in_config_file = FALSE;
	set->ex.tree.unknown = NULL;
	set->ex.tree.selection_check_func = selection_check_func;
	set->ex.tree.draggable_check_func = draggable_check_func;
	e2_option_register (set);
	return set;
}

/**
 * @brief register unknown tree option
 *
 * read @a buf until the end of the tree option and register it
 * with a GList helper as unknown option.
 *
 * @param name string with the name of the option
 * @param i    current line counter
 * @param buf  NULL-terminated array of strings in config file format
 *
 * @return     new line counter
 **/
gint e2_option_tree_register_unknown (gchar *name, gint i, gchar *buf[])
{
	printd (DEBUG, "e2_option_tree_register_unknown (name:%s,i:%d,buf:)", name, i);
	gint save = i;
	gchar *line;
	GList *list = NULL;
	while ((line = buf[i++]) != NULL)
	{
		list = g_list_append (list, g_strdup (line));
		if (g_str_equal (line, ">"))
		{
			gchar **strv = e2_list_to_strv (list);
			g_list_free (list);
			e2_option_unknown_register2 (name, strv);
			return (i - 1);
		}
	}
	//failure, this doesn't look like a tree option
	return save;
}

E2_OptionTreeSet *e2_option_tree_add_column (E2_OptionSet *set, gchar *name,
	E2_OptionTreeType type, gint idef, gchar *sdef, void *extra,
	void *extra_data, E2_OptionTreeSetFlags flags, gpointer visible_check_func,
	gpointer visible_check_func_data)
{
	E2_OptionTreeSet *col = malloc (sizeof (E2_OptionTreeSet));
	col->name = name;
	col->type = type;
	col->idef = idef;
	col->sdef = sdef;
	col->extra = extra;
	col->extra_data = extra_data;
	set->ex.tree.columns_num++;
	set->ex.tree.columns = g_list_append (set->ex.tree.columns, col);
	col->flags = flags;
	col->visible_check_func = visible_check_func;
	col->visible_check_func_data = visible_check_func_data;
	return col;
}

void e2_option_tree_sync_columns (E2_OptionSet *set)
{
	GType t[set->ex.tree.columns_num + 1];
	GList *column;
	gint i = 0;
	for (column = set->ex.tree.columns; column != NULL; column = g_list_next (column))
	{
		E2_OptionTreeSet *set = column->data;
		switch (set->type)
		{
			case E2_OPTION_TREE_TYPE_BOOL:
				t[i] = G_TYPE_INT;
				break;
			case E2_OPTION_TREE_TYPE_STR:
			case E2_OPTION_TREE_TYPE_ICON:
			case E2_OPTION_TREE_TYPE_SEL:
			case E2_OPTION_TREE_TYPE_KEY:
				t[i] = G_TYPE_STRING;
				break;
			default:
				printd (ERROR, "internal error, list options may only be bool,"
					" string or int");
				break;
		}
		i++;
	} 
	//column for visible check
	t[set->ex.tree.columns_num] = G_TYPE_BOOLEAN;
	gtk_tree_store_set_column_types (set->ex.tree.model, set->ex.tree.columns_num + 1, t); 
	//check if there are values for this option that have been read from the
	//config file before
	if (set->ex.tree.unknown != NULL)
	{
		e2_option_read_from_string (set->ex.tree.unknown);
		g_strfreev (set->ex.tree.unknown);
		set->ex.tree.unknown = NULL;
	}
	
}

void e2_option_tree_revert (gchar *option_name)
{
	E2_OptionSet *set = _e2_option_tree_check_helper (option_name);
	_e2_option_tree_revert (set, FALSE);
}

void _e2_option_tree_revert (E2_OptionSet *set, gboolean empty_check)
{
	if (!(set->ex.tree.found_in_config_file && empty_check) &&
		(set->ex.tree.def != NULL))
	{
		
		g_object_unref (set->ex.tree.model);
		set->ex.tree.model = gtk_tree_store_new (1, G_TYPE_INT);
		e2_option_tree_sync_columns (set);
		//option needs default configuration which is stored in ex.tree.def
		//of the OptionSet
		e2_option_read_from_string (set->ex.tree.def);
	}
}

void _e2_option_tree_append_default (E2_OptionSet *set, gchar *first, ...)
{
	if (first == NULL) return;
	va_list args;
	va_start (args, first);
	gchar *cur = first;
	GList *list = NULL;
	while (cur)
	{
		list = g_list_append (list, cur);
		cur = va_arg (args, gchar *);
	}
	va_end (args);
	set->ex.tree.def_num = g_list_length (list);
	gchar **lines = malloc ((set->ex.tree.def_num + 1) * sizeof (gpointer));
	gint i;
	for (i = 0; i < set->ex.tree.def_num; i++)
	{
		lines[i] = list->data;
		list = list->next;
	}
	lines[set->ex.tree.def_num] = NULL;
	set->ex.tree.def = lines;
	
//	g_list_foreach (list + 1), (void *) g_free, NULL);
	g_list_free (list);
}

E2_OptionSet *_e2_option_tree_check_helper (gchar *option)
{
	E2_OptionSet *set;
	set = (E2_OptionSet *) g_hash_table_lookup (options_hash, option);
	if (set == NULL)
	{
		printd (WARN, "trying to add to tree '%s' which doesn't exist", option);
		return NULL;
	}
	if (set->type != E2_OPTION_TYPE_TREE)
	{
		printd (WARN, "trying to add to tree '%s' which isn't a tree", set->name);
		return NULL;
	}
	return set;
}

void e2_option_tree_add (gchar *option, GtkTreeIter *iter, gint n_options,
	void *options[])
{
	E2_OptionSet *set = _e2_option_tree_check_helper (option);
	if (set != NULL)
		_e2_option_tree_add (set, iter, NULL, FALSE, FALSE, n_options, options);
}

void _e2_option_tree_add (E2_OptionSet *set, GtkTreeIter *iter, GtkTreeIter *parent,
	gboolean sibling, gboolean before, gint n_options, void *options[])
{
	if (sibling == TRUE)
	{
		if (before == TRUE)
			gtk_tree_store_insert_before (set->ex.tree.model, iter, NULL, parent);
		else
			gtk_tree_store_insert_after (set->ex.tree.model, iter, NULL, parent);
	}
	else
	{
		if (before == TRUE)
			gtk_tree_store_insert_before (set->ex.tree.model, iter, parent, NULL);
		else
			gtk_tree_store_insert_after (set->ex.tree.model, iter, parent, NULL);
	}

	gtk_tree_store_set (set->ex.tree.model, iter, set->ex.tree.columns_num, TRUE, -1);

	gint i;
	for (i = 0; i < n_options; i++)
	{
		if (options[i] != NULL)
			gtk_tree_store_set (set->ex.tree.model, iter, i, options[i], -1);
	}
}

void e2_option_tree_add_default (gchar *option, GtkTreeIter *iter, GtkTreeIter *parent, gboolean sibling)
{
	E2_OptionSet *set = _e2_option_tree_check_helper (option);
	if (set != NULL)
	_e2_option_tree_add_default (set, iter, parent, sibling);
}

void _e2_option_tree_add_default (E2_OptionSet *set, GtkTreeIter *iter, GtkTreeIter *parent, gboolean sibling)
{
	void *arg[set->ex.tree.columns_num];
	GList *column;
	gint i = 0;
	for (column = set->ex.tree.columns; column != NULL; column = g_list_next (column))
	{
		E2_OptionTreeSet *opt = column->data;
		switch (opt->type)
		{
			case E2_OPTION_TREE_TYPE_BOOL:
				arg[i] = (void *) opt->idef;
				break;
			case E2_OPTION_TREE_TYPE_SEL:
			case E2_OPTION_TREE_TYPE_STR:
			case E2_OPTION_TREE_TYPE_ICON:
			case E2_OPTION_TREE_TYPE_KEY:
				arg[i] = opt->sdef;
				break;
			default:
				break;
		}
		i++;
	}
	
	_e2_option_tree_add (set, iter, parent, sibling, FALSE, set->ex.tree.columns_num, arg);

	if ((parent != NULL) && (sibling == FALSE))
		gtk_tree_view_expand_all (GTK_TREE_VIEW (set->widget));
	gtk_tree_view_set_cursor  (GTK_TREE_VIEW (set->widget),
		gtk_tree_model_get_path (GTK_TREE_MODEL (set->ex.tree.model), iter),
		gtk_tree_view_get_column (GTK_TREE_VIEW (set->widget), 0), FALSE);

	if (set->widget)
	{
		gtk_widget_grab_focus (set->widget);
	}
}

void e2_option_tree_del (gchar *option, GtkTreeIter *iter)
{
	E2_OptionSet *set = _e2_option_tree_check_helper (option);
	if (set != NULL)
		_e2_option_tree_del (set, iter);
}

void _e2_option_tree_del (E2_OptionSet *set, GtkTreeIter *iter)
{
	gtk_tree_store_remove (set->ex.tree.model, iter);
}

void _e2_option_tree_move (E2_OptionSet *set, GtkTreeIter *iter, GtkTreePath *path_dest, gboolean sibling, gboolean before)
{
	//find out real number of columns, we also want to move the meta data
	gint n = gtk_tree_model_get_n_columns (set->ex.tree.model);
	gpointer values[n];
	gint i;
	//save old row contents
	for (i = 0; i < n; i++)
	{
		gtk_tree_model_get (set->ex.tree.model, iter, i, values + i, -1);
	}
	GtkTreePath *path_old = gtk_tree_model_get_path (set->ex.tree.model, iter);
	GtkTreeRowReference *ref_old = gtk_tree_row_reference_new (set->ex.tree.model, path_old);
	GtkTreeIter iter2;
	if ((gtk_tree_model_get_iter (set->ex.tree.model, &iter2, path_dest)) == TRUE)
	{
		GtkTreeIter iter_new;
		_e2_option_tree_add (set, &iter_new, &iter2, sibling, before, n, values);

		gtk_tree_path_free (path_old);
		path_old = gtk_tree_row_reference_get_path (ref_old);
		if ((gtk_tree_model_get_iter (set->ex.tree.model, &iter2, path_old)) == TRUE)
		{
			if (gtk_tree_model_iter_has_child (set->ex.tree.model, &iter2))
			{
				gboolean expand = FALSE;
				if (gtk_tree_view_row_expanded (GTK_TREE_VIEW (set->widget), path_old))
				{
					expand = TRUE;
					gtk_tree_view_collapse_row (GTK_TREE_VIEW (set->widget), path_old);
				}
				
				GtkTreePath *path_new = gtk_tree_model_get_path (set->ex.tree.model, &iter_new);
				GtkTreeIter iter_child;
				while (gtk_tree_model_iter_children (set->ex.tree.model, &iter_child, &iter2))
					_e2_option_tree_move (set, &iter_child, path_new, FALSE, TRUE);

				if (expand)
					gtk_tree_view_expand_row (GTK_TREE_VIEW (set->widget), path_new, FALSE);
				gtk_tree_path_free (path_new);
			}
			_e2_option_tree_del (set, &iter2);
		}

		gtk_tree_view_set_cursor  (GTK_TREE_VIEW (set->widget),
			path_dest, gtk_tree_view_get_column (GTK_TREE_VIEW (set->widget), 0), FALSE);
	}
	gtk_tree_path_free (path_old);
	gtk_tree_row_reference_free (ref_old);
}

void e2_option_tree_set (gchar *option, GtkTreeIter *iter, gint col, void *data)
{
}
	
void _e2_option_tree_set (E2_OptionSet *set, GtkTreeIter *iter, gint col, void *data)
{
	gtk_tree_store_set (GTK_TREE_STORE (set->ex.tree.model), iter, col, data, -1);
}

void *_e2_option_tree_get_void (E2_OptionTreeSet *set, gchar *option)
{
	switch (set->type)
	{
		case E2_OPTION_TREE_TYPE_BOOL:
			if (g_str_equal (option, "true")) return (void *) TRUE;
			else return (void *) FALSE;
			break;
		case E2_OPTION_TREE_TYPE_STR:
		case E2_OPTION_TREE_TYPE_ICON:
		case E2_OPTION_TREE_TYPE_SEL:
		case E2_OPTION_TREE_TYPE_KEY:
			return (void *) option;
		default:
			return NULL;
			break;
	}
	return NULL;
}

void _e2_option_tree_write_to_file (FILE *f, E2_OptionSet *set,
	GtkTreeIter *iter, gint level)
{
	while (TRUE)
	{
		gint j;
		for (j = 0; j < level; j++)
			g_fprintf (f, "\t");
		GList *column;
		j = 0;
		for (column = set->ex.tree.columns; column != NULL; column = g_list_next (column))
		{
			E2_OptionTreeSet *opt = column->data;
			switch (opt->type)
			{
				case E2_OPTION_TREE_TYPE_BOOL:
					{
						gint int_data;
						gtk_tree_model_get (set->ex.tree.model, iter,  j, &int_data, -1);
						if (int_data) g_fprintf (f, "true");
						else g_fprintf (f, "false");
					}
					break;
				case E2_OPTION_TREE_TYPE_STR:
				case E2_OPTION_TREE_TYPE_ICON:
				case E2_OPTION_TREE_TYPE_SEL:
				case E2_OPTION_TREE_TYPE_KEY:
					{
						gchar *str_data;
						gtk_tree_model_get (set->ex.tree.model, iter,  j, &str_data, -1);
						
						if (str_data != NULL)
						{
						if ((j == 0) && ((str_data[0] == '<') || (str_data[0] == '>')))
							g_fprintf (f, "\\");
						if ((str_data[0] != '\0') && ((strchr (str_data, '|')) != NULL))
						{
							gchar **split = g_strsplit (str_data, "|", -1);
							g_free (str_data);
							str_data = g_strjoinv ("\\|", split);
							g_strfreev (split);
						}
						}
						g_fprintf (f, "%s", str_data != NULL ? str_data : "");
   						g_free (str_data);
					}
					break;
				default:
					break;
			}
			g_fprintf (f, "|");
			j++;
		}
		fseek (f, -1L, SEEK_CUR);
		g_fprintf (f, "\n");

		if ((gtk_tree_model_iter_has_child (set->ex.tree.model, iter)) == TRUE)
		{
			GtkTreeIter iter2;
			gtk_tree_model_iter_children (set->ex.tree.model, &iter2, iter);
			_e2_option_tree_write_to_file (f, set, &iter2, level + 1);
		}
		if ((gtk_tree_model_iter_next (set->ex.tree.model, iter)) != TRUE)
			return;
	}
}

