/*
 * $Id: fileops.c,v 1.3 2004/08/08 22:38:14 cairo Exp $
 *
 * Copyright (C) 2003 tooar <tooar@gmx.net>
 * Copyright (C) 1999 Michael Clark.
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
//#include <errno.h>
//#include <pthread.h>
#include "emelfm2.h"
#include "debug.h"
#include "e2_option.h"
#include "e2_button.h"
#include "e2_dialog.h"
#include "e2_command.h"

/*
 * This file contains functions used to handle different file operations
 * like copying, deleting, sym linking, etc.. most functions do this by 
 * forking and running the standard unix commands for the appropriate operation
 */

/* copy src to dest.. return when the operation is done */
gint file_copy(gchar *src, gchar *dest)
{
	gchar *quote1 = g_shell_quote (src);
	gchar *quote2 = g_shell_quote (dest);
	gchar *command = g_strconcat ("cp -rf ", quote1, " ", quote2, NULL);
	e2_command_run (command, E2_COMMAND_RANGE_FILE_OPERATION);
	g_free (command);
	g_free (quote1);
	g_free (quote2);
	return 0;
}

/* delete file.. return when the operation is done */
gint file_delete(gchar *file)
{
	gchar *quote = g_shell_quote (file);
	gchar *command = g_strconcat ("rm -rf ", quote, NULL);
	e2_command_run (command, E2_COMMAND_RANGE_FILE_OPERATION);
	g_free (command);
	g_free (quote);
	return 0;
}

/* move src to dest.. return when the operation is done */
gint file_move(gchar *src, gchar *dest)
{
	gchar *quote1 = g_shell_quote (src);
	gchar *quote2 = g_shell_quote (dest);
	gchar *command = g_strconcat ("mv -f ", quote1, " ", quote2, NULL);
	e2_command_run (command, E2_COMMAND_RANGE_FILE_OPERATION);
	g_free (command);
	g_free (quote1);
	g_free (quote2);
	return 0;
}

gint file_chmod(gchar *path, gchar *mode, gboolean recurse_dirs)
{
	gchar *quote = g_shell_quote (path);
	gchar *command = g_strconcat ("chmod", recurse_dirs ? " -R " : " ",
		mode, " ", quote, NULL);
	e2_command_run (command, E2_COMMAND_RANGE_FILE_OPERATION);
	g_free (command);
	g_free (quote);
	return 0;
}

gint file_chown(gchar *path, uid_t owner_id,
	gid_t group_id, gboolean recurse_dirs)
{
	gchar *quote = g_shell_quote (path);
	gchar *mode = g_strdup_printf ("%d:%d", (gint) owner_id, (gint) group_id);
	gchar *command = g_strconcat ("chown", recurse_dirs ? " -R " : " ",
		mode, " ", quote, NULL);
	e2_command_run (command, E2_COMMAND_RANGE_FILE_OPERATION);
	g_free (mode);
	g_free (command);
	g_free (quote);
	return 0;
}
	
/* symlink src to dest */
gint file_symlink(gchar *src, gchar *dest)
{
	gchar *quote1 = g_shell_quote (src);
	gchar *quote2 = g_shell_quote (dest);
	gchar *command = g_strconcat ("ln -s ", quote1, " ", quote2, NULL);
	e2_command_run (command, E2_COMMAND_RANGE_FILE_OPERATION);
	g_free (command);
	g_free (quote1);
	g_free (quote2);
	return 0;
}

gint file_mkdir(gchar *path)
{
	return mkdir(path, 0777);
}

gint op_with_ow_check (gchar *src, gchar *dest, gint (*op_func) (gchar *, gchar *))
{
	printd (DEBUG, "overwrite check");
	
	gchar *s = g_path_get_basename (src);
	gchar *utf = _STR2UTF (s);
	gchar *mesc = MESC (utf);
	g_free (utf);
	gchar *dialog_text = g_strconcat  ("are you sure you want to overwrite <b><u>",
		mesc, "</u></b>?", NULL);
	g_free (mesc);
	GtkWidget *dialog = e2_dialog_create (GTK_STOCK_DIALOG_QUESTION,
		dialog_text, "confirm overwrite", NULL, NULL);
	gint result = e2_dialog_show (dialog, app.main_window,
		E2_BUTTON_CANCEL | E2_BUTTON_NO |
		E2_BUTTON_YES_TO_ALL | E2_BUTTON_YES,
		E2_DIALOG_RUN | E2_DIALOG_FREE);
	g_free (dialog_text);
	gchar *quote1 = g_shell_quote (src);
	gchar *quote2 = g_shell_quote (dest);
	gint retval = -1;
	switch (result)
	{
		case GTK_RESPONSE_YES:
			printd (DEBUG, "overwrite_check: yes");
			op_func(src, dest);
			retval = YES;
			break;
		//111 is yes to all
		case 111:
			printd (DEBUG, "overwrite_check: yes to all");
			op_func(src, dest);
			retval = YES_TO_ALL;
			break;
		case GTK_RESPONSE_NO:
			printd (DEBUG, "overwrite_check: no");
			retval = NO;
			break;
		default:
			printd (DEBUG, "overwrite_check: cancel");
			retval = CANCEL;
	}
	g_free (quote1);
	g_free (quote2);
	
	return retval;
}
