/* $Id: e2_widget.c,v 1.10 2004/07/14 23:20:05 cairo Exp $
 *
 * Copyright (C) 2003-2004 tooar <tooar@gmx.net>
 * Copyright (C) 1999 Michael Clark.
 *
 * This file is part of emelFM2.
 * emelFM2 is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * emelFM2 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with emelFM2; see the file COPYING. If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/**
 * @file src/utils/e2_widget.c 
 * @brief miscelleanous GtkWidget utility functions
 *
 * This file contains various miscelleanous GtkWidget utility functions.
 **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "emelfm2.h"
#include "e2_widget.h"

void e2_widget_set_tooltip (GtkTooltips *tooltips,
	GtkWidget *widget, gchar *text)
{
	if ((text == NULL) || (*text == '\0'))
		return;
	//use default tooltip group
	if (tooltips == NULL)
		tooltips = app.tooltips;
	gtk_tooltips_set_tip  (tooltips, widget, text, NULL);
}

GtkWidget *e2_widget_get_icon (gchar *icon, GtkIconSize size)
{
	if ((icon == NULL) || (*icon == '\0'))
		return NULL;

	GtkWidget *image = NULL;
	GtkStockItem item;
	if ((gtk_stock_lookup (icon, &item)) ||
		(g_str_equal (icon, GTK_STOCK_MISSING_IMAGE)))
		image = gtk_image_new_from_stock (icon, size);
	else
	{
		GdkPixbuf *pixbuf;
		if ((pixbuf = gdk_pixbuf_new_from_file (icon, NULL)) != NULL)
		{
			GdkPixbuf *pixbuf2;
			gint width, height;
			if (!gtk_icon_size_lookup (size, &width, &height))
			{
				width = 16; height = 16;
			}
			pixbuf2 = gdk_pixbuf_scale_simple (pixbuf, width, height,
				GDK_INTERP_BILINEAR);
			image = gtk_image_new_from_pixbuf (pixbuf2);
			g_object_unref (pixbuf2);
			g_object_unref (pixbuf);

		}
	}
	return image;
}

GtkWidget *e2_widget_add_label (GtkWidget *box, gchar *text, gfloat align,
	gint fill, gint pad)
{
	return _e2_widget_add_label (box, text, align, 0.5, fill, pad);
}

GtkWidget *_e2_widget_add_label (GtkWidget *box, gchar *text,
	gfloat xalign, gfloat yalign, gint fill, gint pad)
{
	GtkWidget *label = gtk_label_new (text);
	if (box != NULL)
		gtk_box_pack_start (GTK_BOX (box), label, fill, TRUE, pad);
	gtk_label_set_use_markup (GTK_LABEL (label), TRUE);
	gtk_misc_set_alignment (GTK_MISC (label), xalign, yalign);
	gtk_label_set_line_wrap (GTK_LABEL (label), TRUE);
	gtk_label_set_selectable (GTK_LABEL (label), FALSE);
	gtk_widget_show (label);
	return label;
}

GtkWidget *e2_widget_add_entry (GtkWidget *box, gchar *init_text,
	gint fill, gboolean select_text)
{
	GtkWidget *entry;
	entry = gtk_entry_new ();
	if (init_text != NULL)
		gtk_entry_set_text (GTK_ENTRY (entry), init_text);
	if (select_text)
		gtk_editable_select_region (GTK_EDITABLE (entry), 0, -1);
	gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);
	if (box != NULL)
		gtk_box_pack_start (GTK_BOX (box), entry, fill, TRUE, 5);
	gtk_widget_show (entry);

	return entry;
}

GtkWidget *e2_widget_get_box (gboolean vertical, gboolean homogen, gint spacing)
{
	GtkWidget *box;
	if (vertical)
		box = gtk_vbox_new (homogen, spacing);
	else
		box = gtk_hbox_new (homogen, spacing);
	return box;
}

GtkWidget *e2_widget_add_box (GtkWidget *parent, gint fill, gint pad,
	gboolean vertical, gboolean homogen, gint spacing)
{
	GtkWidget *box = e2_widget_get_box (vertical, homogen, spacing);
	gtk_box_pack_start (GTK_BOX (parent), box, fill, TRUE, pad);
	gtk_widget_show (box);
	return box;
}

/**
 * @brief  create a GtkScrolledWindow
 *
 * This function creates a GtkScrolledWindow. The parameters @a h_policy and
 * @a v_policy are used to set the scrolling policy of the scrolled window.
 * The shadow type is set to GTK_SHADOW_NONE.
 *
 * @param h_policy  scrolled window horizontal scrolling policy
 * @param v_policy  scrolled window vertical scrolling policy
 *
 * @return  the scrolled window
 **/
GtkWidget *e2_widget_get_sw (GtkPolicyType h_policy,
	GtkPolicyType v_policy)
{
	GtkScrolledWindow *scrolled_window = GTK_SCROLLED_WINDOW
		(gtk_scrolled_window_new (NULL, NULL));
	gtk_scrolled_window_set_policy (scrolled_window, h_policy, v_policy);
	gtk_scrolled_window_set_shadow_type (scrolled_window, GTK_SHADOW_NONE);
	return GTK_WIDGET (scrolled_window);
}

/**
 * @brief  create a GtkScrolledWindow and add it to a GtkBox
 *
 * This function creates a GtkScrolledWindow using e2_widget_get_sw() and adds
 * it to @a box. The parameters @a h_policy and @a v_policy are used as
 * arguements for e2_widget_get_sw(). @a fill and @a pad are used as arguments for
 * gtk_box_pack_start() to adjust the packing of the scrolled window.
 *
 * @param box  the parent box to pack the scrolled window to
 * @param h_policy  scrolled window horizontal scrolling policy
 * @param v_policy  scrolled window vertical scrolling policy
 * @param fill  what to do with additional free space in the box
 * @param pad  number of pixels to pad the scrolled window in the box  
 *
 * @return  the scrolled window
 **/
GtkWidget *e2_widget_add_sw (GtkWidget *box, GtkPolicyType h_policy,
	GtkPolicyType v_policy, gint fill, gint pad)
{
	GtkWidget *scrolled_window = e2_widget_get_sw (h_policy, v_policy);
	gtk_box_pack_start (GTK_BOX (box), scrolled_window, fill, TRUE, pad);
	gtk_widget_show (scrolled_window);
	return scrolled_window;
}

/**
 * @brief  add a child with a viewport to a GtkScrolledWindow
 *
 * In contrast to the @a e2_widget_[add|get]* functions, this function does not
 * create a scrolled window. It only adds a viewport to the scrolled window
 * @a sw and then adds @a child to the viewport. his is done by calling
 * gtk_scrolled_window_add_with_viewport(). The shadow type of the viewport is
 * set to GTK_SHADOW_NONE.
 *
 * @param sw  the scrolled window
 * @param child  the child that should be added to the scrolled window
 *
 * @return
 **/
void e2_widget_sw_add_with_viewport (GtkWidget *sw, GtkWidget *child)
{
	gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (sw), child);
	gtk_viewport_set_shadow_type (GTK_VIEWPORT (GTK_BIN (sw)->child),
		GTK_SHADOW_NONE);
}

GtkWidget *add_separator (GtkWidget *box, gint fill, gint pad)
{
	GtkWidget *sep;
	if (GTK_IS_HBOX (box))
	{
		sep = gtk_vseparator_new ();
		gtk_box_pack_start (GTK_BOX (box), sep, fill, TRUE, pad);
	} else if (GTK_IS_VBOX (box))
	{
		sep = gtk_hseparator_new ();
		gtk_box_pack_start (GTK_BOX (box), sep, fill, TRUE, pad);
	} else
		return NULL;
	gtk_widget_show (sep);
	return sep;
}

GtkWidget *add_table (GtkWidget      *box,
		      gint           rows,
		      gint           cols,
		      gint           homogen,
		      gint           fill,
		      gint           pad)
{
	GtkWidget *table;
	table = gtk_table_new (rows, cols, homogen);
	gtk_box_pack_start (GTK_BOX (box), table, fill, TRUE, pad);
	gtk_widget_show (table);

	return table;
}

GtkWidget *add_label_to_table (GtkWidget           *table,
		               gchar               *text,
		               gfloat              align,
		               gint left, gint right, gint top, gint bottom)
{
	GtkWidget *label;
	label = gtk_label_new (text);
	gtk_misc_set_alignment (GTK_MISC (label), align, 0.5);
	gtk_table_attach_defaults (GTK_TABLE (table), label, left, right, top, bottom);
	gtk_widget_show (label);

	return label;
}

GtkWidget *add_entry_to_table (GtkWidget         *table,
		               gchar             *init_text,
		               gint left, gint right, gint top, gint bottom)
{
	GtkWidget *entry;
	entry = gtk_entry_new ();
	if (init_text != NULL)
		gtk_entry_set_text (GTK_ENTRY (entry), init_text);
	gtk_table_attach_defaults (GTK_TABLE (table), entry, left, right, top, bottom);
	gtk_widget_show (entry);

	return entry;
}

GtkWidget *add_button_to_table (GtkWidget       *table,
		                gchar           *label,
		                void *func,
		                gpointer        data,
		                gint left, gint right, gint top, gint bottom)
{
	GtkWidget *button;
	button = gtk_button_new_with_label (label);
	if (func != NULL)
		g_signal_connect (GTK_OBJECT (button), "clicked", G_CALLBACK (func), data);
	gtk_table_attach_defaults (GTK_TABLE (table), button, left, right, top, bottom);
	gtk_widget_show (button);

	return button;
}

GtkWidget *e2_button_add_toggle_to_table (GtkWidget       *table,
		                      gchar           *label,
		                      gint            state,
		                      void *func,
		                      gpointer        data,
		                      gint left, gint right, gint top, gint bottom)
{
	GtkWidget *check_button;
	check_button = gtk_check_button_new_with_label (label);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (check_button), state);
	if (func != NULL)
		g_signal_connect (GTK_OBJECT (check_button), "toggled", G_CALLBACK (func), data);
	gtk_table_attach_defaults (GTK_TABLE (table), check_button,
		                        left, right, top, bottom);
	gtk_widget_show (check_button);

	return check_button;
}

GtkWidget *add_framed_table (GtkWidget *box, gchar *title, gint rows, gint cols,
	gint fill, gint pad)
{
	GtkWidget *frame, *table;

	frame = gtk_frame_new (title);
	gtk_box_pack_start (GTK_BOX (box), frame, fill, TRUE, pad);
	gtk_widget_show (frame);

	table = gtk_table_new (rows, cols, FALSE);
	gtk_container_set_border_width (GTK_CONTAINER (table), 2);
	gtk_table_set_row_spacings (GTK_TABLE (table), 5);
	gtk_container_add (GTK_CONTAINER (frame), table);
	gtk_widget_show (table);

	return table;
}

GtkWidget *add_framed_widget (GtkWidget *box, gchar *title, GtkWidget *widget,
	gint fill, gint pad)
{
	GtkWidget *frame;

	frame = gtk_frame_new (title);
	gtk_box_pack_start (GTK_BOX (box), frame, fill, TRUE, pad);
	gtk_container_add (GTK_CONTAINER (frame), widget);
	gtk_widget_show (widget);
	gtk_widget_show (frame);

	return frame;
}

GtkWidget *e2_widget_add_frame (GtkWidget *box, gboolean fill, gint pad,
	gchar *title, gboolean strech)
{
	gchar *real_title = NULL;
	if (title != NULL)
		real_title = g_strconcat (" ", title, " ", NULL);
	GtkWidget *frame = gtk_frame_new (real_title);
	if (title != NULL)
	{
		g_free (real_title);
		GtkWidget *label = gtk_frame_get_label_widget (GTK_FRAME (frame));
		gtk_label_set_use_markup (GTK_LABEL (label), TRUE);
	}
	gtk_box_pack_start (GTK_BOX (box), frame, fill, fill, pad);
	return frame;
}

void e2_widget_set_cursor (GdkCursorType type)
{
	GdkCursor *cursor = gdk_cursor_new (type);
	gdk_window_set_cursor (app.main_window->window, cursor);
	gdk_cursor_unref (cursor);
	gdk_flush ();
}

GtkWidget *e2_widget_add_eventbox (GtkWidget *box, gboolean fill, gint pad)
{
	GtkWidget *event = gtk_event_box_new ();
	gtk_box_pack_start (GTK_BOX (box), event, fill, fill, pad);
	return event;
}

/**
 * @brief create a GtkNotebook
 *
 * This function create a GtkNotebook and connects @a func to the
 * "switch-patch" signal of the GtkNotebook.
 *
 * @param func  the switch-patch callback function or NULL
 * @param data  data pointer for the callback function
 *
 * @return  a new GtkNotebook
 **/
GtkWidget *e2_widget_get_notebook (gpointer func, gpointer data)
{
	GtkWidget *notebook = gtk_notebook_new ();
	if (func != NULL)
		g_signal_connect (notebook, "switch-page", G_CALLBACK (func), data);
	return notebook;

}

/**
 * @brief  create a notebook and add it to a box
 *
 * This function creates a new GtkNotebook using e2_widget_get_notebook() and
 * adds it to @a box.
 * 
 * @param box  the box to add the notebook to
 * @param fill  fill parameter for the box child
 * @param pad  pad parameter for the box child
 * @param func  the switch-patch callback function for the notebook or NULL
 * @param data  data pointer for the callback function
 *
 * @return  the notebook that has been added to @a box
 **/
GtkWidget *e2_widget_add_notebook (GtkWidget *box, gboolean fill, gint pad,
	gpointer func, gpointer data)
{
	GtkWidget *notebook = e2_widget_get_notebook (func, data);
	gtk_box_pack_start (GTK_BOX (box), notebook, fill, fill, pad);
	return notebook;
}

/**
 * @brief  get the size of the font
 *
 * This function retrieves approximate character width and height for the
 * font currently assigned to the widget
 * 
 * @param widget  the widget using the font
 * @param width  address of int to accept the character width in pixels
 * @param height  address of int to accept the character height inpixels
 *
 **/
void e2_widget_get_font_pixels (GtkWidget *widget, gint *width, gint *height)
{
	PangoContext *context = gtk_widget_get_pango_context (widget);
	PangoFontMetrics *metrics = pango_context_get_metrics (context,
			widget->style->font_desc, pango_context_get_language (context));
	*width = PANGO_PIXELS (pango_font_metrics_get_approximate_char_width
		(metrics));
	*height = PANGO_PIXELS (pango_font_metrics_get_ascent (metrics) +
				pango_font_metrics_get_descent (metrics));
	pango_font_metrics_unref (metrics);
}


/**
 * @brief  set a font for a widget
 *
 * This function sets widget font to "font_string, optionally retrieving
 * approximate pixel character width and height, pixel sizes are
 * not retrieved if either width or height are NULL.
 * 
 * @param widget  widget using the font
 * @param font_string  string describing the font
 * @param width  address to accept the int pixel width, or NULL if un-needed
 * @param height  address to accept the int pixel height, or NULL if un-needed 
 *
 **/
void e2_widget_set_font (GtkWidget *widget, gchar *font_string, gint *width,
	gint *height)
{
	PangoFontDescription *font_desc;
	font_desc = pango_font_description_from_string (font_string);
	gtk_widget_modify_font (widget, font_desc);
	pango_font_description_free (font_desc);
	if ((width == NULL) || (height == NULL))
		return;
	e2_widget_get_font_pixels (widget, width, height);
}

